/*
 * Copyright (c) 2019 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-BSD-5-Clause-Nordic
 */

#include <stdio.h>
#include <string.h>

#include <zephyr.h>
#include <modem/at_cmd.h>
#include <net/coap.h>
#include <modem/at_notif.h>
#include <net/socket.h>
#include <errno.h>
#include <fcntl.h>
#include <modem/lte_lc.h>
#if defined(CONFIG_LWM2M_CARRIER)
#include <lwm2m_carrier.h>
#endif

#define APP_COAP_SEND_INTERVAL_MS 5000
#define APP_COAP_MAX_MSG_LEN 1280
#define APP_COAP_VERSION 1

static int sock;
static struct pollfd fds;
static struct sockaddr_storage server;
static u16_t next_token;

static u8_t coap_buf[APP_COAP_MAX_MSG_LEN];

LOG_MODULE_REGISTER(dt_modem_at_cmd, 4);	
#if defined(CONFIG_BSD_LIBRARY)

/**@brief Recoverable BSD library error. */
void bsd_recoverable_error_handler(uint32_t err)
{
	printk("bsdlib recoverable error: %u\n", (unsigned int)err);
}

#endif /* defined(CONFIG_BSD_LIBRARY) */

/**@brief Resolves the configured hostname. */
static int server_resolve(void)
{
	int err;
	struct addrinfo *result;
	struct addrinfo hints = {
		.ai_family = AF_INET,
		.ai_socktype = SOCK_DGRAM
	};
	char ipv4_addr[NET_IPV4_ADDR_LEN];

	err = getaddrinfo(CONFIG_COAP_SERVER_HOSTNAME, NULL, &hints, &result);
	if (err != 0) {
		printk("ERROR: getaddrinfo failed %d\n", err);
		return -EIO;
	}

	if (result == NULL) {
		printk("ERROR: Address not found\n");
		return -ENOENT;
	}

	/* IPv4 Address. */
	struct sockaddr_in *server4 = ((struct sockaddr_in *)&server);

	server4->sin_addr.s_addr =
		((struct sockaddr_in *)result->ai_addr)->sin_addr.s_addr;
	server4->sin_family = AF_INET;
	server4->sin_port = htons(CONFIG_COAP_SERVER_PORT);

	inet_ntop(AF_INET, &server4->sin_addr.s_addr, ipv4_addr,
		  sizeof(ipv4_addr));
	printk("IPv4 Address found %s\n", ipv4_addr);

	/* Free the address. */
	freeaddrinfo(result);

	return 0;
}

/**@brief Initialize the CoAP client */
static int client_init(void)
{
	int err;

	sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock < 0) {
		printk("Failed to create CoAP socket: %d.\n", errno);
		return -errno;
	}

	err = connect(sock, (struct sockaddr *)&server,
		      sizeof(struct sockaddr_in));
	if (err < 0) {
		printk("Connect failed : %d\n", errno);
		return -errno;
	}
	
    int fl = fcntl(sock, F_GETFL, 0);
	err = fcntl(sock, F_SETFL, fl | O_NONBLOCK);
    if(err < 0) {
		printk("fnctl O_NONBLOCK failed : %d.", errno);
		return -errno;
    }

	/* Initialize FDS, for poll. */
	fds.fd = sock;
	fds.events = POLLIN;

	/* Randomize token. */
	next_token = sys_rand32_get();

	return 0;
}

/**@brief Handles responses from the remote CoAP server. */
static int client_handle_get_response(u8_t *buf, int received)
{
	int err;
	struct coap_packet reply;
	const u8_t *payload;
	u16_t payload_len;
	u8_t token[8];
	u16_t token_len;
	u8_t temp_buf[16];

	err = coap_packet_parse(&reply, buf, received, NULL, 0);
	if (err < 0) {
		printk("Malformed response received: %d\n", err);
		return err;
	}

	payload = coap_packet_get_payload(&reply, &payload_len);
	token_len = coap_header_get_token(&reply, token);

	if ((token_len != sizeof(next_token)) &&
	    (memcmp(&next_token, token, sizeof(next_token)) != 0)) {
		printk("Invalid token received: 0x%02x%02x\n",
		       token[1], token[0]);
		return 0;
	}

	snprintf(temp_buf, MAX(payload_len, sizeof(temp_buf)), "%s", payload);

	printk("CoAP response: code: 0x%x, token 0x%02x%02x, payload: %s\n",
	       coap_header_get_code(&reply), token[1], token[0], temp_buf);

	return 0;
}

/**@biref Send CoAP GET request. */
static int client_get_send(void)
{
	int err;
	struct coap_packet request;

	next_token++;

	err = coap_packet_init(&request, coap_buf, sizeof(coap_buf),
			       APP_COAP_VERSION, COAP_TYPE_NON_CON,
			       sizeof(next_token), (u8_t *)&next_token,
			       COAP_METHOD_GET, coap_next_id());
	if (err < 0) {
		printk("Failed to create CoAP request, %d\n", err);
		return err;
	}

	err = coap_packet_append_option(&request, COAP_OPTION_URI_PATH,
					(u8_t *)CONFIG_COAP_RESOURCE,
					strlen(CONFIG_COAP_RESOURCE));
	if (err < 0) {
		printk("Failed to encode CoAP option, %d\n", err);
		return err;
	}

	err = send(sock, request.data, request.offset, 0);
	if (err < 0) {
		printk("Failed to send CoAP request, %d\n", errno);
		return -errno;
	}

	printk("CoAP request sent: token 0x%04x\n", next_token);

	return 0;
}

#if defined(CONFIG_LWM2M_CARRIER)
K_SEM_DEFINE(carrier_registered, 0, 1);

void lwm2m_carrier_event_handler(const lwm2m_carrier_event_t *event)
{
	switch (event->type) {
	case LWM2M_CARRIER_EVENT_BSDLIB_INIT:
		printk("LWM2M_CARRIER_EVENT_BSDLIB_INIT\n");
		break;
	case LWM2M_CARRIER_EVENT_CONNECT:
		printk("LWM2M_CARRIER_EVENT_CONNECT\n");
		break;
	case LWM2M_CARRIER_EVENT_DISCONNECT:
		printk("LWM2M_CARRIER_EVENT_DISCONNECT\n");
		break;
	case LWM2M_CARRIER_EVENT_READY:
		printk("LWM2M_CARRIER_EVENT_READY\n");
		k_sem_give(&carrier_registered);
		break;
	case LWM2M_CARRIER_EVENT_FOTA_START:
		printk("LWM2M_CARRIER_EVENT_FOTA_START\n");
		break;
	case LWM2M_CARRIER_EVENT_REBOOT:
		printk("LWM2M_CARRIER_EVENT_REBOOT\n");
		break;
	}
}
#endif /* defined(CONFIG_LWM2M_CARRIER) */

/**@brief Configures modem to provide LTE link. Blocks until link is
 * successfully established.
 */
static void modem_configure(void)
{
#if defined(CONFIG_LTE_LINK_CONTROL)
	if (IS_ENABLED(CONFIG_LTE_AUTO_INIT_AND_CONNECT)) {
		/* Do nothing, modem is already turned on
		 * and connected.
		 */
	} else {
#if defined(CONFIG_LWM2M_CARRIER)
		/* Wait for the LWM2M_CARRIER to configure the modem and
		 * start the connection.
		 */
		printk("Waitng for carrier registration...\n");
		k_sem_take(&carrier_registered, K_FOREVER);
		printk("Registered!\n");
#else /* defined(CONFIG_LWM2M_CARRIER) */
		int err;

		printk("LTE Link Connecting ...\n");
		err = lte_lc_init_and_connect();
		__ASSERT(err == 0, "LTE link could not be established.");
		printk("LTE Link Connected!\n");
#endif /* defined(CONFIG_LWM2M_CARRIER) */
	}
#endif /* defined(CONFIG_LTE_LINK_CONTROL) */
}

/* Returns 0 if data is available.
 * Returns -EAGAIN if timeout occured and there is no data.
 * Returns other, negative error code in case of poll error.
 */
static int wait(int timeout)
{
	int ret = poll(&fds, 1, timeout);

	if (ret < 0) {
		printk("poll error: %d\n", errno);
		return -errno;
	}

	if (ret == 0) {
		/* Timeout. */
		return -EAGAIN;
	}

	if ((fds.revents & POLLERR) == POLLERR) {
		printk("wait: POLLERR\n");
		return -EIO;
	}

	if ((fds.revents & POLLNVAL) == POLLNVAL) {
		printk("wait: POLLNVAL\n");
		return -EBADF;
	}

	if ((fds.revents & POLLIN) != POLLIN) {
		return -EAGAIN;
	}

	return 0;
}


static void dt_modem_at_handler(void *context, const char *response)
{
	int res;
    if (response == NULL)
    {
        LOG_ERR("Response buffer is NULL-pointer");
        return;
    }
    printk("Processing notification: %s", response);
}

void main(void)
{
	s64_t next_msg_time = APP_COAP_SEND_INTERVAL_MS;
	int err, received;
	
	
	int ret = at_notif_register_handler(NULL, dt_modem_at_handler);
	if (ret != 0)
	{
		LOG_ERR("Failed to register handler [%d]", ret);
		return -EIO;
	}

	printk("The nRF CoAP client sample started\n");

	modem_configure();
	
	
  // enable network time support
  if (at_cmd_write("AT%XNETTIME=1", NULL, 0, NULL) != 0)
  {
    LOG_ERR("Failed to enable network time support");
    return -5;
  }
  LOG_DBG("Enabled network time support");

  // subscribe to signal quality notifications
  if (at_cmd_write("AT%CESQ=1", NULL, 0, NULL) != 0)
  {
    LOG_ERR("Failed to subscribe to signal quality notifications (CESQ=1");
    return -5;
  }
  LOG_DBG("Subscribed to signal quality notifications (CESQ=1)");

  // subscribe to SNR notifications
  if (at_cmd_write("AT%XSNRSQ=1", NULL, 0, NULL) != 0)
  {
    LOG_ERR("Failed to subscribe to SNR notifications (XSNRSQ=1");
    return -6;
  }
  LOG_DBG("Subscribed to SNR notifications (XSNRSQ=1)");

  // subscribe to network status notifications
  if (at_cmd_write("AT+CEREG=5", NULL, 0, NULL) != 0)
  {
    LOG_ERR("Failed to subscribe to network status notifications (CEREG=5");
    return -7;
  }
  LOG_DBG("Subscribed to network status notifications (CEREG=5)");

  // subscribe to CGEV notifications
  if (at_cmd_write("AT+CGEREP=1", NULL, 0, NULL) != 0)
  {
    LOG_ERR("Failed to subscribe to CGEV notifications (CGEREP=1");
    return -8;
  }
  LOG_DBG("Subscribed to CGEV notifications (CGEREP=1)");

  // subscribe to network time notifications
  if (at_cmd_write("AT%XTIME=1", NULL, 0, NULL) != 0)
  {
    LOG_ERR("Failed to subscribe to network time notifications (XTIME=1");
    return -9;
  }
  LOG_DBG("Subscribed to network time notifications (XTIME=1)");

  
  // subscribe to connection status notifications
  if (at_cmd_write("AT+CSCON=1", NULL, 0, NULL) != 0)
  {
    LOG_ERR("Failed to subscribe to connection status notifications (CSCON=1");
    return -10;
  }
  LOG_DBG("Subscribed to connection status notifications (CSCON=1)");


	if (server_resolve() != 0) {
		printk("Failed to resolve server name\n");
		return;
	}

	if (client_init() != 0) {
		printk("Failed to initialize CoAP client\n");
		return;
	}

	next_msg_time = k_uptime_get();

	while (1) {
		for(int i = 0; i < 20; i++) {
			client_get_send();
			k_yield();
			k_sleep(K_SECONDS(3));
		}
		printk("Waiting for notifications\n");
		k_yield();
		k_sleep(K_SECONDS(60));
	}

	(void)close(sock);
}
